/**
 *  MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm.rdb;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.metainfo.rdb.api.RdbClassMapping;
import org.objectweb.jorm.runtime.ConnectionSpecJDBC;
import org.objectweb.jorm.runtime.MapperJDBC;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.optim.jorm.JormLeafRewriter;
import org.objectweb.medor.optim.jorm.RdbJormLeafRewriter;
import org.objectweb.medor.optim.jorm.TestJormQueryRewriter;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.query.lib.BasicOrderField;

import java.util.ArrayList;

/**
 * @author S.Chassande-Barrioz
 */
public class TestRdbJormQueryRewriter extends TestJormQueryRewriter {

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestRdbJormQueryRewriter.class);
	}

	public TestRdbJormQueryRewriter() {
		super("TestRdbJormQueryRewriter", Log.MEDOR_PREFIX + ".optim.jorm.rdb.queryrewriter");
	}

	public TestRdbJormQueryRewriter(String testName) {
		super(testName, Log.MEDOR_PREFIX + ".optim.jorm.rdb.queryrewriter");
	}

	public PMapper newMappper() {
		PMapper m = new MapperJDBC();
		try {
			ConnectionSpecJDBC cs = new ConnectionSpecJDBC("jdbc:product:basename", "java.lang.String");
			m.setConnectionFactory(cs);
		} catch (PException e) {
			e.printStackTrace();
		}
		return m;
	}

	public QueryRewriter createQueryRewriter() {
		ArrayList rules = new ArrayList(2);
		rules.add(new FlattenQueryTreeRule());
		rules.add(new JormFlatten2Rdb());
		return new BasicQueryRewriter(rules);
	}

	public QueryLeaf getQueryLeafA(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_REF_SING_A_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "id", PTypeSpace.LONG, "id_col", qt);
		RdbExpField f1 = leaf.addRdbField(prefix + "field1", PTypeSpace.STRING, "FIELD_1", qt);
		leaf.removeRdbField(f1);
		leaf.setQueryFilter(new Equal(
				new BasicFieldOperand(f1),
				new BasicParameterOperand(PTypeSpace.STRING, "?1")));
		return leaf;
	}

	public QueryLeaf getQueryLeafOrderByA(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_REF_SING_A_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "id", PTypeSpace.LONG, "id_col", qt);
		RdbExpField f1 = leaf.addRdbField(prefix + "field1", PTypeSpace.STRING, "FIELD_1", qt);
		leaf.removeRdbField(f1);
		RdbExpField f2 = leaf.addRdbField(prefix + "field2", PTypeSpace.STRING, "FIELD_2", qt);
		leaf.removeRdbField(f2);
		OrderField of1 = new BasicOrderField(f1, true);
		OrderField of2 = new BasicOrderField(f2);
		leaf.setOrderBy(new OrderField[] {of1, of2});
		leaf.setQueryFilter(new Equal(
				new BasicFieldOperand(f1),
				new BasicParameterOperand(PTypeSpace.STRING, "?1")));
		return leaf;
	}

	public QueryLeaf getQueryLeafMemberOfBD(Example eB, Example eD) throws MedorException {
		RdbClassMapping cmB = (RdbClassMapping)
				eB.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		RdbClassMapping cmD = (RdbClassMapping)
				eD.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = eB.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qtB = new BasicQualifiedTable(cmB.getRdbTable().getName(), "T_B_0");
		BasicRdbExpQueryLeaf leafB = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtB}, "");
		RdbExpField fc2 = leafB.addRdbField(prefix + "c2", PTypeSpace.CHAR, "C2", qtB);
		RdbExpField fc1 = leafB.addRdbField(prefix + "c1", PTypeSpace.BYTE, "C1", qtB);

		QualifiedTable qtD = new BasicQualifiedTable(cmD.getRdbTable().getName(), "T_D_1");
		BasicRdbExpQueryLeaf leafD = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtD}, "");
		RdbExpField fb = leafD.addRdbField(prefix + "b", PTypeSpace.BYTE, "B", qtD);
		RdbExpField fa = leafD.addRdbField(prefix + "a", PTypeSpace.CHAR, "A", qtD);

		ArrayList left = new ArrayList(2);
		ArrayList right = new ArrayList(2);
		left.add(new BasicFieldOperand(fc2));
		left.add(new BasicFieldOperand(fc1));
		right.add(new BasicFieldOperand(fb));
		right.add(new BasicFieldOperand(fa));
		MemberOf mo = new MemberOf(left, right);
		leafB.setQueryFilter(mo);
		return leafB;
	}

	public QueryLeaf getQueryLeafMemberOfBC(Example eB, Example eC) throws MedorException {
		RdbClassMapping cmB = (RdbClassMapping)
				eB.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		RdbClassMapping cmC = (RdbClassMapping)
				eC.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = eB.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qtB = new BasicQualifiedTable(cmB.getRdbTable().getName(), "T_B_0");
		BasicRdbExpQueryLeaf leafB = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtB}, "");
		RdbExpField fc2 = leafB.addRdbField(prefix + "c2", PTypeSpace.CHAR, "C2", qtB);
		RdbExpField fc1 = leafB.addRdbField(prefix + "c1", PTypeSpace.BYTE, "C1", qtB);
		RdbExpField fy = leafB.addRdbField(prefix + "y", PTypeSpace.LONG, "Y", qtB);
		leafB.removeRdbField(fy);

		QualifiedTable qtC = new BasicQualifiedTable(cmC.getRdbTable().getName(), "T_C_1");
		BasicRdbExpQueryLeaf leafC = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtC}, "");
		RdbExpField fx = leafC.addRdbField(prefix + "x", PTypeSpace.LONG, "X", qtC);
		leafC.setQueryFilter(new Equal(new BasicFieldOperand(fx), new BasicOperand((long) 2)));
		ArrayList left = new ArrayList(1);
		ArrayList right = new ArrayList(1);
		left.add(new BasicFieldOperand(fy));
		right.add(new BasicFieldOperand(fx));
		MemberOf mo = new MemberOf(left, right);
		leafB.setQueryFilter(mo);
		return leafB;
	}
}
