/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.filter.TestExpressionHelper;
import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.PropagFromNestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbStringQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;

/**
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class TestQueryTreeHelper extends TestMedorHelper {

	TestQueryTreeHelper(String testName) {
		super(testName);
	}

	public static void equals(String prefix, QueryTree qt1, QueryTree qt2,
							  Logger log)
		throws MedorException {
		if (qt1 == null && qt2 == null)
			return;
		if (qt1 == null ^ qt2 == null)
			fail(prefix + "one of the query tree instances is null: qt1=" + qt1
				+ " / qt2=" + qt2);
		if (qt1.getDistinct() != qt2.getDistinct())
			fail(prefix + "Not same distinct qt1=" + qt1.getDistinct()
				+ " /qt2=" + qt2.getDistinct());
		if (qt1.getClass() != qt2.getClass())
			fail(prefix + "Not same class qt1=" + qt1.getClass().getName()
				+ " / qt2=" + qt2.getClass().getName());

		log.log(BasicLevel.DEBUG, "QueryNode: name=" + qt1.getName() + " / class: " + qt1.getClass().getName());

		equals(prefix, qt1.getTupleStructure(), qt2.getTupleStructure(), log);

		if (qt1 instanceof QueryNode) {
			QueryNode qn1 = (QueryNode) qt1;
			QueryNode qn2 = (QueryNode) qt2;
			TestExpressionHelper.equals(prefix,
				qn1.getQueryFilter(), qn2.getQueryFilter(), log);
			assertEquals(prefix + "QueryNode(" + qt1.getName() + "): Not same type", qn1.getType(), qn2.getType());
			assertEquals(prefix + "QueryNode(" + qt1.getName() + "): Number of child is different",
				qn1.getChildren().length, qn2.getChildren().length);
			ArrayList child1 = QueryTreePrinter.getChildren(qn1);
			ArrayList child2 = QueryTreePrinter.getChildren(qn2);
			for (int i = 0; i < child1.size(); i++) {
				equals(prefix,
					(QueryTree) child1.get(i),
					(QueryTree) child2.get(i), log);
			}
		}
		else if (qt1 instanceof ClassExtent) {
			ClassExtent je1 = (ClassExtent) qt1;
			ClassExtent je2 = (ClassExtent) qt2;
			assertEquals(prefix + "ClassExtent(" + qt1.getName() + "): Not same meta object",
				((Class) je1.getMetaObject()).getName(), ((Class) je2.getMetaObject()).getName());
			if (je1.getDataStore() != null)
				assertEquals(prefix + "ClassExtent(" + qt1.getName() + "): Not same datstore",
					je1.getDataStore().getName(), je2.getDataStore().getName());
		}
		else if (qt1 instanceof GenClassExtent) {
			GenClassExtent je1 = (GenClassExtent) qt1;
			GenClassExtent je2 = (GenClassExtent) qt2;
			assertEquals(prefix + "ClassExtent(" + qt1.getName() + "): Not same meta object",
				((GenClassRef) je1.getMetaObject()).getName(), ((GenClassRef) je2.getMetaObject()).getName());
			assertEquals(prefix + "ClassExtent(" + qt1.getName() + "): Not same mapper",
				je1.getPMapper(), je2.getPMapper());
			assertEquals(prefix + "ClassExtent(" + qt1.getName() + "): Not same datstore",
				je1.getDataStore(), je2.getDataStore());
		}
		else if (qt1 instanceof RdbExpQueryLeaf) {
			RdbExpQueryLeaf ql1 = (RdbExpQueryLeaf) qt1;
			RdbExpQueryLeaf ql2 = (RdbExpQueryLeaf) qt2;
			QualifiedTable[] t1 = ql1.getQualifiedTables();
			QualifiedTable[] t2 = ql2.getQualifiedTables();
			assertEquals(prefix + "RdbExpQueryLeaf(" + qt1.getName() + "): Number of QualifiedTable is different",
				t1.length, t2.length);
			for (int i = 0; i < t1.length; i++) {
				assertEquals(prefix + "RdbExpQueryLeaf(" + qt1.getName() + "): Not same table name",
					t1[i].getTableName(), t2[i].getTableName());
				assertEquals(prefix + "RdbExpQueryLeaf(" + qt1.getName() + "): Not same alias name",
					t1[i].getAliasName(), t2[i].getAliasName());
			}
			//TODO check SQL request
			assertEquals(prefix + "Not same SQL request", ql1.getSqlRequest(null), ql2.getSqlRequest(null));
		}
		else if (qt1 instanceof RdbStringQueryLeaf) {
			RdbStringQueryLeaf ql1 = (RdbStringQueryLeaf) qt1;
			RdbStringQueryLeaf ql2 = (RdbStringQueryLeaf) qt2;
			try {
				assertEquals(prefix + "RdbStringQueryLeaf(" + qt1.getName() + "): Not same request",
					ql1.getSqlRequest(null), ql2.getSqlRequest(null));
			}
			catch (MedorException e) {
				log.log(BasicLevel.ERROR,
					prefix + "RdbStringQueryLeaf(" + qt1.getName() + "): Impossible to fetch the sqlRequest: ",
					e);
				throw e;
			}
		}
	}

	public static void equals(String prefix,
							  TupleStructure ts1,
							  TupleStructure ts2,
							  Logger log)
		throws MedorException {
		Field[] f1s = ts1.getFields();
		Field[] f2s = ts2.getFields();
		assertEquals(prefix + "TupleStructure: not same size",
			f1s.length, f2s.length);
		for (int i = 0; i < f1s.length; i++) {
			equals(prefix, f1s[i], f2s[i], log);
		}
	}

	public static void equals(String prefix, Field f1, Field f2, Logger log)
		throws MedorException {
		if (f1 == null && f2 == null)
			return;
		if (f1 == null ^ f2 == null)
			fail(prefix + "one of the fields is null: f1=" + f1
				+ " / f2=" + f2);
		if (f1.getClass() != f2.getClass())
			fail(prefix + "Not same class f1=" + f1.getClass().getName()
				+ " / f2=" + f2.getClass().getName());
		log.log(BasicLevel.DEBUG, "Field class: " + f1.getClass().getName());

		assertEquals(prefix + "Field: Name is different", f1.getName(), f2.getName());
		log.log(BasicLevel.DEBUG, "Field name: " + f1.getName());

		if (f1.getType() != null) {
			log.log(BasicLevel.DEBUG, "Field type f1: " + f1.getType().getJavaName() + " / " + f1.getType().getJormName());
			log.log(BasicLevel.DEBUG, "Field type f2: " + f2.getType().getJavaName() + " / " + f2.getType().getJormName());
		}
		assertTrue(prefix + "Field: (" + f1.getName() + ") Type is different (was " + f2.getType().getTypeCode() + ", expected " + f1.getType().getTypeCode(),
			f1.getType().isa(f2.getType()));
		log.log(BasicLevel.DEBUG, "Field type: " + f1.getType());

		if (f1 instanceof CalculatedField) {
			TestExpressionHelper.equals(prefix,
				((CalculatedField) f1).getExpression(),
				((CalculatedField) f2).getExpression(),
				log);
		}
		else if (f1 instanceof PropagatedField) {
			Field[] fs1 = ((PropagatedField) f1).getPreviousFields();
			Field[] fs2 = ((PropagatedField) f2).getPreviousFields();
			assertEquals(prefix +
				"PropagatedField: Number of previous field is different",
				fs1.length, fs2.length);
			if (f1 instanceof PropagFromNestedField) {
				fs1 = ((PropagFromNestedField) f1).getPreviousFlatField();
				fs2 = ((PropagFromNestedField) f2).getPreviousFlatField();
				assertEquals(prefix +
					"PropagatedField: Number of previous field is different",
					fs1.length, fs2.length);
			}
		}
		else if (f1 instanceof RdbExpField) {
			RdbExpField ref1 = (RdbExpField) f1;
			RdbExpField ref2 = (RdbExpField) f2;
			assertEquals("RdbExpField: Not same column name",
				ref1.getColumnName(), ref2.getColumnName());
			assertEquals("RdbExpField: Not same table name",
				ref1.getTable().getTableName(), ref2.getTable().getTableName());
			assertEquals("RdbExpField: Not same table name",
				ref1.getTable().getAliasName(), ref2.getTable().getAliasName());
		}
		else if (f1 instanceof RdbField) {
			RdbField ref1 = (RdbField) f1;
			RdbField ref2 = (RdbField) f2;
			assertEquals("RdbField: Not same column name",
				ref1.getColumnName(), ref2.getColumnName());
		}
		else if (f1 instanceof JormField) {
			log.log(BasicLevel.DEBUG, "Compare JormField: " + f1.getName()
				+ " qt=" + ((QueryTreeField) f1).getQueryTree().getName());
			log.log(BasicLevel.DEBUG, "with JormField: " + f2.getName()
				+ " qt=" + ((QueryTreeField) f2).getQueryTree().getName());
			assertEquals(
				prefix + "JormField: not same TypedElement",
				((JormField) f1).getTypedElement(),
				((JormField) f2).getTypedElement());
		}
		else if (f1 instanceof PNameField) {
			PNameField pnf1 = (PNameField) f1;
			PNameField pnf2 = (PNameField) f2;
			if (pnf1.isInGenClass())
				assertEquals(prefix + "PNameField: not same genclassRef",
					pnf1.getGenClassRef().getName(), pnf2.getGenClassRef().getName());
			if (pnf1.getMetaObjectClass() != null)
				assertEquals(prefix + "PNameField: not same MetaObjectClass",
					pnf1.getMetaObjectClass().getName(), pnf2.getMetaObjectClass().getName());
			if (pnf1.getReference() != null)
				assertEquals(prefix + "PNameField: not same reference",
					pnf1.getReference().getName(), pnf2.getReference().getName());
			assertEquals(prefix + "PNameField: not same isClassPName",
				pnf1.isClassPName(), pnf2.isClassPName());
			assertEquals(prefix + "PNameField: not same isInGenClass",
				pnf1.isInGenClass(), pnf2.isInGenClass());
		}
		else
			log.log(BasicLevel.WARN,
				"Unknown field type: " + f1.getClass().getName());
	}
}
