/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.query.jorm;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.NavigatorOperator;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.Or;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.jorm.TestJormHelper;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class TestNavigatorNodeFactory extends TestJormPDHelper {

    public TestNavigatorNodeFactory(String testName, String loggerName) {
        super(testName, loggerName);
    }


    public NavigatorOperator getNavigatorOperator(Field f, String[] path)
		throws MedorException {
        if (path.length == 0)
            return null;
        NavigatorOperator nav =
            new NavigatorOperator(new BasicFieldOperand(f), path[0]);
        for (int i = 1; i < path.length; i++) {
            nav = new NavigatorOperator(nav, path[i]);
        }
        return nav;
    }

    /**
	 * Test the NavigatorNodeFactory.navigate method.
	 * - start with a pname of the class
	 * - d.f1 where f1 is a primitive field.
	 * ==> The result is the same QueryTree
	 */
	public void _test1a() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", e.extent, result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.navigate method.
	 * - start with a pname of the class
	 * - d.e where e is a reference field.
	 * ==> The result is the same QueryTree.
	 */
    public void _test1b() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", e.extent, result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.navigate method.
	 * - start with a pname of the class
	 * - d.es where es is a GenClass of E.
	 * The result is a join with the GC node.
	 */
    public void _test1c() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
			logger.log(BasicLevel.DEBUG, "Rewritten QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);

            PNameField pnf = (PNameField)
                e.extent.getTupleStructure().getField(e.clazz.getFQName() + ".setOfE");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent =
                new GenClassExtent(gcr, e.clazz.getFQName() + ".setOfE", "object", NavigatorNodeFactory.GENCLASS_ELEMENT_NAME);

            JoinProject join = new JoinProject("", null);

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".setOfE");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(
                    fToAdd[i].getName(),
                    fToAdd[i].getType(),
                    new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
            }

            //project in 'join' the element field of the GenClassExtent
			Field elemToAdd = extent.getField(extent.getElementFieldName());
			join.addPropagatedField(elemToAdd.getName(), elemToAdd.getType(),
									new QueryTreeField[]{(QueryTreeField) elemToAdd});

            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );
			logger.log(BasicLevel.DEBUG, "Expected QueryTree");
			QueryTreePrinter.printQueryTree(join, logger);

            TestQueryTreeHelper.equals("", join, result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.navigate/IS_EMPTY method.
	 * - start with a pname of the class
	 * - d.es where es is a GenClass of E.
	 * - add isEmpty to the path traversal
	 * The result is a join with the GC node.
	 */
    public void _test1cEmpty() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result1, result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result1 = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.IS_EMPTY).getQueryTree();
			logger.log(BasicLevel.DEBUG, "QueryTree with IsEmpty");
            QueryTreePrinter.printQueryTree(result1, logger);

			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			//assign mapper to all extents
			Iterator extents = JormQueryTreeHelper.getJormExtents(result1).iterator();
			while (extents.hasNext()) {
				JormExtent ext = (JormExtent)extents.next();
				logger.log(BasicLevel.DEBUG, "Setting mapper for extent " + ext.getName());
				ext.setPMapper(mapper, PROJECT_NAME);
				ext.setProjectName("test");
			}
			result = jqr.transform(result1);
			logger.log(BasicLevel.DEBUG, "Rewritten QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
			//TODO add expected query result

/*
            PNameField pnf = (PNameField)
                e.extent.getTupleStructure().getField(e.clazz.getFQName() + ".setOfE");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent =
                new GenClassExtent(gcr, e.clazz.getFQName() + ".setOfE", "object", NavigatorNodeFactory.GENCLASS_ELEMENT_NAME);

            JoinProject join = new JoinProject("", null);

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".setOfE");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(
                    fToAdd[i].getName(),
                    fToAdd[i].getType(),
                    new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
            }

            //project in 'join' the element field of the GenClassExtent
			Field elemToAdd = extent.getField(extent.getElementFieldName());
			join.addPropagatedField(elemToAdd.getName(), elemToAdd.getType(),
									new QueryTreeField[]{(QueryTreeField) elemToAdd});

            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );
			logger.log(BasicLevel.DEBUG, "Expected QueryTree");
			QueryTreePrinter.printQueryTree(join, logger);

            TestQueryTreeHelper.equals("", join, result, logger);
		*/
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    public void _test1d() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, "test1d",
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
            PNameField pnf = (PNameField)
                e.extent.getTupleStructure().getField(e.clazz.getFQName() + ".setOfE");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent =
            new GenClassExtent(gcr, e.clazz.getFQName() + ".setOfE", "object", NavigatorNodeFactory.GENCLASS_ELEMENT_NAME);

            JoinProject join = new JoinProject("", null);

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".setOfE");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(
                    fToAdd[i].getName(),
                    fToAdd[i].getType(),
                    new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
            }

            //project in 'join' all fields of the GenClassExtent
            // and find the QueryTreeField which represents the element
			Field elemToAdd = extent.getField(extent.getElementFieldName());
			join.addPropagatedField("test1d", elemToAdd.getType(),
									new QueryTreeField[]{(QueryTreeField) elemToAdd});

			join.setQueryFilter(
                new Equal(
                    new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)
                )
            );

            TestQueryTreeHelper.equals("", join, result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.navigate method.
	 * - start with a field of the class
	 * - d.f1 where f1 is a primitive field.
	 * ==> The result is a join with the D node
	 */
	public void _test2a() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        try {
            QueryTreeField f =
				(QueryTreeField) e.extent.getField(e.clazz.getFQName() + ".d");
            String[] path = new String[]{"f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION).getQueryTree();
			logger.log(BasicLevel.DEBUG, "Result tree of test2a");
            QueryTreePrinter.printQueryTree(result, logger);
			logger.log(BasicLevel.DEBUG, "Expected tree of test2a");
            QueryTreePrinter.printQueryTree(get2abExpected(e, "f1"), logger);
            TestQueryTreeHelper.equals("", get2abExpected(e, "f1"), result,
				logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

	/**
	 * Test the NavigatorNodeFactory.navigate method.
	 * - start with a field of the class
	 * - d.e where e is a reference field.
	 * ==> The result is a join with the D node
	 */
    public void _test2b() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(e.clazz.getFQName() + ".d");
            String[] path = new String[]{"e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", get2abExpected(e, "e"), result,
				logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    public QueryTree get2abExpected(Example e, String fn)
		throws MedorException {
        Example ed = getExtentOfD(new String[]{fn});
        ClassExtent extent = ed.extent;
        JoinProject join = new JoinProject("", null);

        //project in 'join' all fields of the inherited node
        Field[] fToAdd = e.extent.getFields();
        for (int i = 0; i < fToAdd.length; i++) {
            join.addPropagatedField(
                fToAdd[i].getName(),
                fToAdd[i].getType(),
                new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
        }

		fToAdd = extent.getFields();
        for (int i = 0; i < fToAdd.length; i++) {
            join.addPropagatedField(
                fToAdd[i].getName(),
                fToAdd[i].getType(),
                new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
        }

        join.setQueryFilter(
            new Equal(
                new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() +".d")),
                new BasicFieldOperand(extent.getField(ed.clazz.getFQName() + ".object"))
            )
        );
        return join;
    }

	/**
	 * Test the NavigatorNodeFactory.navigate method.
	 * - start with a field of the class
	 * - d.es where es is a GenClass of E.
	 * The result is a join(join(qt, node(D), node(GC)).
	 */
	public void _test2c() {
		TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
		try {
			Field f = e.extent.getField(e.clazz.getFQName() + ".d");
			String[] path = new String[]{"setOfE"};
			NavigatorOperator nav = getNavigatorOperator(f, path);
			result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
			QueryTreePrinter.printQueryTree(result, logger);
		}
		catch (MedorException me) {
			me.printStackTrace();
            fail(me.getMessage());
		}
	}

	//A, b.c.d.f1
	public void _test3a() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
	//A, b.c.d.e
	public void _test3b() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
	//A, [A.b, c, d, es]
	public void _test3c() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);

        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}

	//A, [A.b, c, d, es] ISEMPTY + OR
	public void test3cEmpty() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result, result1 = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result1 = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.IS_NOT_EMPTY).getQueryTree();

			//constructing filter with adding "OR f1=12"
			QueryNode r1 = (QueryNode) result1;
			Field f1 = r1.getTupleStructure().getField(
                getFieldName(e.extent.getName(), "f1"));
			Field prevF1 = ((PropagatedField)f1).getPreviousFields()[0];
			r1.setQueryFilter(
					new Or(r1.getQueryFilter(),
						new Equal(new BasicFieldOperand(prevF1),
								new BasicVariableOperand(12)
						))
			);

			logger.log(BasicLevel.DEBUG, "QueryTree with IsEmpty");
            QueryTreePrinter.printQueryTree(result1, logger);

			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			//assign mapper to all extents
			Iterator extents = JormQueryTreeHelper.getJormExtents(result1).iterator();
			while (extents.hasNext()) {
				JormExtent ext = (JormExtent)extents.next();
				logger.log(BasicLevel.DEBUG, "Setting mapper for extent " + ext.getName());
				ext.setPMapper(mapper, PROJECT_NAME);
				ext.setProjectName("test");
			}
			result = jqr.transform(result1);
			logger.log(BasicLevel.DEBUG, "Rewritten QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
			//TODO add expected query result

        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}

	public void _test3cMemberOf() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree input = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), e.pnameFieldName));
            String[] path = new String[]{e.clazz.getFQName() + ".b", "c", "d", "setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            input = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
			//TODO add MemberOf
            QueryTreePrinter.printQueryTree(input, logger);
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			//assign mapper to all extents
			Iterator extents = JormQueryTreeHelper.getJormExtents(input).iterator();
			while (extents.hasNext()) {
				JormExtent ext = (JormExtent)extents.next();
				logger.log(BasicLevel.DEBUG, "Setting mapper for extent " + ext.getName());
				ext.setPMapper(mapper, PROJECT_NAME);
				ext.setProjectName("test");
			}
			QueryTree result = jqr.transform(input);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}

    //A, [c.d.es]
	public void _test4a() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), "b"));
            String[] path = new String[]{"c", "d", "f1"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
    //A, [c.d.es]
	public void _test4b() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), "b"));
            String[] path = new String[]{"c", "d", "e"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
    //A, [c.d.es]
	public void _test4c() {
		TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(
                getFieldName(e.extent.getName(), "b"));
            String[] path = new String[]{"c", "d", "setOfE"};
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
					NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            QueryTreePrinter.printQueryTree(result, logger);
        }
        catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
	}
}
